/* -----------------------------------------------------------------------------
 *    ___   __    __       _      __              __   __
 *   / _ | / /__ / /____ _| | /| / /__  ___  ____/ /  / /____ ____
 *  / __ |/  '_//  '_/ // / |/ |/ / _ `/ -_)/ __/ _ \/ __/ -_) __/
 * /_/ |_/_/\_\/_/\_\\_,_/|__/|__/\_,_/\__/ \__/_//_/\__/\__/_/
 *
 * 080824-1 (c)2009 Elektor International Media B.V.
 * -------------------------------------------------------------------------- */
/* -----------------------------------------------------------------------------
 * Purpose
 *   Main program file and source code
 *
 * Changes Log
 *  - 1.0 : initial source code
 *
 * Notes
 *  Nothing special, may come later...
 * -------------------------------------------------------------------------- */

/* -----------------------------------------------------------------------------
 * Include
 * -------------------------------------------------------------------------- */
#include "LPC2103.h"
#include "HD44780.h"
#include "common.h"
#include "tools.h"



/* -----------------------------------------------------------------------------
 * CONFIGURATION Defines
 * -------------------------------------------------------------------------- */
// firmware version number
#define VERSION "1.2"



/* -----------------------------------------------------------------------------
 * Defines
 * -------------------------------------------------------------------------- */
// push button is on GPIO P0.2 (active low)
#define PUSH_BUTTON_PIN 0x00000004

// the LCD_BACKLIGHT_MAX_BRIGHTNESS correspond to the PWM duty cicle on the LED [0..100]
#define LCD_BACKLIGHT_MAX_BRIGHTNESS 50

/* Calculate the coefficient to multiply to the value delivered by the ADC
 * to get the input voltage in V.
 * This avoid Floating Point usage at runtime
 */
// ADC is 10 bits wide, set here which value correspond to the refence voltage
#define ADC_MAX_VALUE   1023
// reference voltage of the ADC bloc in milivolts
#define ADC_REF_VOLTAGE 3300
// battery voltage measurement bridge resistor R3 value in kilo Ohms
#define RESISTOR_HIGH    100
// battery voltage measurement bridge resistor R8 nd P1 value in kilo Ohms
#define RESISTOR_LOW    8.53
// the coefficient for voltage measurement
#define COEF_INPUT_VOLTAGE (ADC_REF_VOLTAGE*((RESISTOR_HIGH + RESISTOR_LOW)/RESISTOR_LOW))

/* Calculate the coefficient you need to messure the current,
 * the value will be in nano Ampere.
 * This avoid Floating Point usage at runtime
 */
// ADC is 22 bits wide
#define MCP3550_MAX_VALUE (0x200000 - 1)
// reference voltage on the MCP3550 ADC
#define MCP3550_REF_VOLTAGE  1200000000 //in nV
// shunt resistor value in Ohm for the current measurement
#define RESISTOR_SHUNT       0.05
// the coefficient for current measurement
#define COEF_CURRENT ((MCP3550_REF_VOLTAGE*(1/RESISTOR_SHUNT))/MCP3550_MAX_VALUE)

// number of measured value for the average method
#define SAMPLE_NB_FOR_AVG 32

// maximum battery voltage value
#define BATTERY_MAX_VOLTAGE 14000
// minimum battery voltage value
#define BATTERY_MIN_VOLTAGE 10800
// number of tolerated out of battery voltage value before alert
#define VBAT_MAX_OUTOFRANGE 10

// wordlist for an easy translation
#define TEXT_VOLTAGE_FC  "Voltage  "
#define TEXT_CURRENT_FC  "Current  "
#define TEXT_POWER_FC    "Power    "
#define TEXT_CAPACITY_FC "Capacity "




/* -----------------------------------------------------------------------------
 * dummy definition of the unused function from the startup code
 * -------------------------------------------------------------------------- */
void UNDEF_Routine(){}
void SWI_Routine(){}
void IRQ_Routine(){}
void FIQ_Routine(){} 

void LCD_print(int current){
if (current < 1000){
		//.###
		LCD_print_character('.');
		LCD_int3(current);
	}else if((current > 1000) & (current < 10000)){
		//#.##
		LCD_int1(current/1000);
		LCD_print_character('.');
		LCD_int2((current%1000)/10);
	}else if((current > 10000) & (current < 100000)){
		//##.#
		LCD_int2(current/1000);
		LCD_print_character('.');
		LCD_int1((current%1000)/100);
	}
	else if(current > 100000){
		//####
		LCD_int4(current/1000);
	}
}

/* -----------------------------------------------------------------------------
 * Private Functions
 * -------------------------------------------------------------------------- */
// TODO
/* -------------------------------------------------------------------------- */
/**@author Steffen Graf / Antoine Authier
 * @date   20090306
 * @brief  measure current and voltages and display the values
 */
static void
battery_monitor (void)
{
	static unsigned char uv_error = 0, ov_error = 0;
	unsigned long ad_voltage = 0;
	unsigned long ad_current;
	unsigned long current;
	unsigned short voltage;
	unsigned long capacity_abs;
	static signed long capacity = 0;
	unsigned long power;
	char sign = ' ';
	int i;

	// measure the voltage over the Shunt Resistor
	ad_current = MCP3550_get_value();

	// average the measurement of the battery voltage to reduce noise
	for (i = 0; i < SAMPLE_NB_FOR_AVG; i++) {
		ad_voltage += ADC_get_AD0_value();
	}
	ad_voltage /= SAMPLE_NB_FOR_AVG;

	// calculate the battery voltage (out of the voltage divider)
	voltage = (ad_voltage*COEF_INPUT_VOLTAGE) / 1000;

	// Battery under- and overvoltage detection.
	// Everytime the battery voltage exceeds 13.8V or goes below 10.8V
	// a counter gets increased. If the voltage is out of range for ten seconds,
	// the error LED will light. When the voltage is in range
	// the counter gets reseted and the LED will be turned off.
	if (voltage > BATTERY_MAX_VOLTAGE)
		ov_error++;
	if (voltage < BATTERY_MIN_VOLTAGE)
		uv_error++;
	if (  (ov_error > VBAT_MAX_OUTOFRANGE)
	    ||(uv_error > VBAT_MAX_OUTOFRANGE))
		LED_error_set_status(LED_ON);
	if (  (voltage < BATTERY_MAX_VOLTAGE)
	    &&(ov_error > 0))
	{
		ov_error = 0;
		LED_error_set_status(LED_OFF);
	}
	if (   (voltage > BATTERY_MIN_VOLTAGE)
	    && (uv_error > 0))
	{
		uv_error = 0;
		LED_error_set_status(LED_OFF);
	}

	// clear the LCD before displaying the new values (in random order)
	LCD_clear();

	// check in which direction flows the current in the Shunt Resistor
	// the 21st bit of the MCP3550's value is the sign.
	if (ad_current & (1 << 21))
	{
		// a negative current means charging the battery
		sign = '+';
		// get the correct absolute ad_current value
		ad_current = ~ad_current;
		ad_current &= 0x3FFFFF;
		ad_current++;
	}
	else
	{
		// a positive current means discharging the battery
		sign = '-';
	}
	// calculate the current consumption in mA
	current = (COEF_CURRENT*ad_current) / 1000000;
	// calculate the power dissipation in mW
	power = (voltage*current) / 1000;
	// calculate the new capacity in mAh
	if (sign == '+') {
		capacity += (current * 1000 / 3600);
	} else {
		capacity -= (current * 1000 / 3600);
	}

	// send the voltage value through the serial interface in plain text
	UART0_SendText(TEXT_VOLTAGE_FC": ",EOL_NONE);
	UART0_SendInt5_signed(voltage);
	UART0_SendText("mV", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);

	// display the current consumption
	LCD_move_cursor(LCD_FIRST_LINE, 0);
	LCD_print_character(sign);
	/*
	if (current < 10000)
	{
		LCD_int4(current);
		LCD_print_text("mA  ");
	}
	else
	{
		LCD_int4(current / 1000);
		LCD_print_text("A   ");
	}
	*/
	LCD_print(current);
	LCD_print_text("A   ");
	
	// send the current consumption value throught the serial port in plain text
	UART0_SendText(TEXT_CURRENT_FC": ", EOL_NONE);
	UART0_SendByte(sign);
	UART0_SendInt5(current);
	UART0_SendText("mA", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);

	// display the voltage
	LCD_move_cursor(LCD_FIRST_LINE, 9);
	LCD_int5(voltage);
	LCD_print_text("mV");

	// display the power consumption and send it though the serial interface
	LCD_move_cursor(LCD_SECOND_LINE, 9);
	LCD_print_character(sign);
	UART0_SendText(TEXT_POWER_FC": ", EOL_NONE);
	UART0_SendByte(sign);
	// above 9A display the value in Ampere
	/*if (power > 9000)
	{
		LCD_int4(power / 1000);
		LCD_print_text("W");
		UART0_SendInt5_signed(power);
		UART0_SendText("mW", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);
	}
	else
	{
		LCD_int4(power);
		LCD_print_text("mW");
		UART0_SendInt5(power);
		UART0_SendText("mW", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);
	}*/
	
	LCD_print(power);
	LCD_print_text("W");

	// prepare to display the evaluated capacity of the battery
	if (capacity >= 0)
	{
		sign = '+';
		// get the absolute value to display
		capacity_abs = capacity;
	}
	else
	{
		sign = '-';
		// get the absolute value to display
		capacity_abs = -capacity;
	}
	// display the capacity and send it through the serial port
	LCD_move_cursor(LCD_SECOND_LINE, 0);
	LCD_print_character(sign);
	UART0_SendText(TEXT_CAPACITY_FC": ", EOL_NONE);
	UART0_SendByte(sign);
	/*if (capacity < 9000000)
	{
		LCD_int4(capacity_abs / 1000);
		LCD_print_text("mAh");
		UART0_SendInt5_signed(capacity_abs / 1000);
		UART0_SendText("mAh", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);
	}
	else
	{
		LCD_int4(capacity_abs / 1000000);
		LCD_print_text("Ah");
		UART0_SendInt5_signed(capacity_abs / 1000);
		UART0_SendText("mAh", EOL_LINE_FEED|EOL_CARRIAGE_RETURN);
	}*/
	
	LCD_print(capacity_abs/1000);
	LCD_print_text("Ah");
}


/* -------------------------------------------------------------------------- */
/**@author Steffen Graf
 * @date   20090306
 * @brief  main function
 */
int
main (void)
{
	// set peripheral devices clock (PCLK) to processor clock (CCLK)
	APBDIV = 1;
	// setup hardware timers
	delay_initialize();

	// initialize modules
	LED_error_initialize();
	MCP3550_initilize();
	UART_initialize(BAUDRATE);
	LCD_backlight_init();
	LCD_initialize();

	// display the welcome message
	LCD_move_cursor(LCD_FIRST_LINE, 0);
	LCD_print_text("Akkuwachter v"VERSION);
	LCD_move_cursor(LCD_SECOND_LINE, 0);
	LCD_print_text("by Steffen Graf");
	delay_us(800000);
	// just for a nice user feeling
	LCD_clear();
	delay_us(200000);

	// firmware pooling loop
	while (1) {
		// Checks if the push button is pressed
		if (IOPIN & PUSH_BUTTON_PIN) {
			// if it's not pressed, turn the backlight off
			LCD_backlight_set_cycle(0);
		} else {
			// if it's pressed, turn the backlight on
			LCD_backlight_set_cycle(LCD_BACKLIGHT_MAX_BRIGHTNESS);
		}

		// monitor the battery every seconds
		if (T0TC >= 1000) {
			// reset Timer0 counter
			T0TC = 0;
			// measure and display values
			battery_monitor();
		}
	}
}
